// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1999.
// ---------------------------------------------------------------------------
//	waveOut based driver
// ---------------------------------------------------------------------------
//	$Id: soundwo.cpp,v 1.2 1999/07/04 08:50:00 cisc Exp $

#include "headers.h"
#include "soundwo.h"

using namespace WinSoundDriver;

// ---------------------------------------------------------------------------
//	\zEj ---------------------------------------------------------------

DriverWO::DriverWO()
{
	src = 0;
	playing = false;
	hthread = 0;
	hwo = 0;
	mixalways = false;
}

DriverWO::~DriverWO()
{
	Cleanup();
}

// ---------------------------------------------------------------------------
//   -------------------------------------------------------------------

bool DriverWO::Init(SoundBuffer* s, HWND, uint rate, uint ch, uint buflen)
{
	if (playing)
		return false;

	src = s;
	sampleshift = 1 + (ch == 2 ? 1 : 0);

	// XbhN
	if (!hthread)
	{
		hthread = HANDLE(_beginthreadex(NULL, 0, ThreadEntry,
							reinterpret_cast<LPVOID>(this), 0, &idthread));
		if (!hthread)
			return false;
		SetThreadPriority(hthread, THREAD_PRIORITY_ABOVE_NORMAL);
	}

	// vZ
	buffersize = (rate * ch * sizeof(Sample) * buflen / 1000 / 4) & ~7;

	// ĐtH[}bgݒ
	WAVEFORMATEX wf;
    memset(&wf, 0, sizeof(WAVEFORMATEX));
    wf.wFormatTag = WAVE_FORMAT_PCM;
	wf.nChannels = ch;
    wf.nSamplesPerSec = rate;
	wf.wBitsPerSample = 16;
	wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
	wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;

	if (waveOutOpen(&hwo, WAVE_MAPPER, &wf, idthread, 
					reinterpret_cast<DWORD> (this), CALLBACK_THREAD) 
			!= MMSYSERR_NOERROR)
	{
		hwo = 0;
		return false;
	}
	
	playing = true;
	dontmix = true;
	numblocks = 0;
	for (int i=0; i<4; i++)
		SendBlock();
	dontmix = false;
	return true;
}

// ---------------------------------------------------------------------------
//  Еt -----------------------------------------------------------------

bool DriverWO::Cleanup()
{
	if (playing)
	{
		playing = false;
		if (hwo)
		{
			while (waveOutReset(hwo) == MMSYSERR_HANDLEBUSY)
				Sleep(10);
			while (waveOutClose(hwo) == MMSYSERR_HANDLEBUSY)
				Sleep(10);
			while (numblocks > 0)
				Sleep(10);
			hwo = 0;
		}
	}
	if (hthread)
	{
		PostThreadMessage(idthread, WM_QUIT, 0, 0);
		if (WAIT_TIMEOUT == WaitForSingleObject(hthread, 3000))
			TerminateThread(hthread, 0);
		CloseHandle(hthread);
		hthread = 0;
	}
	return true;
}

// ---------------------------------------------------------------------------
//  ubN1 -----------------------------------------------------------------

bool DriverWO::SendBlock()
{
	if (playing)
	{
		WAVEHDR* whdr = new WAVEHDR;
		if (!whdr)
			return false;
		
		memset(whdr, 0, sizeof(WAVEHDR));
		
		whdr->lpData = new char[buffersize];
		if (whdr->lpData)
		{
			whdr->dwBufferLength = buffersize;
			whdr->dwUser = 0;
			whdr->dwFlags = 0;
			whdr->dwLoops = 0;
			whdr->lpNext = NULL;
			whdr->reserved = 0;

			if (!dontmix && (mixalways || !src->IsEmpty()))
				src->Get((Sample*) whdr->lpData, buffersize >> sampleshift);
			else
				memset(whdr->lpData, 0, buffersize);
				
			CriticalSection::Lock lock(cs);
			if (!waveOutPrepareHeader(hwo, whdr, sizeof(WAVEHDR)))
			{
				if (!waveOutWrite(hwo, whdr, sizeof(WAVEHDR)))
				{
					numblocks ++;
					return true;
				}
				// s
				waveOutUnprepareHeader(hwo, whdr, sizeof(WAVEHDR));
			}
		}
		delete[] whdr->lpData;
		delete whdr;
	}
	return false;
}

// ---------------------------------------------------------------------------
//  ubNĐ̏ -----------------------------------------------------

void DriverWO::BlockDone(WAVEHDR* whdr)
{
	if (whdr)
	{
		CriticalSection::Lock lock(cs);
		waveOutUnprepareHeader(hwo, whdr, sizeof(WAVEHDR));
		delete[] whdr->lpData;
		delete whdr;
		numblocks --;
	}
	SendBlock();
}

// ---------------------------------------------------------------------------
//  Xbh -------------------------------------------------------------

uint __stdcall DriverWO::ThreadEntry(LPVOID arg)
{
	DriverWO* dw = reinterpret_cast<DriverWO*> (arg);
	MSG msg;

	while (GetMessage(&msg, NULL, 0, 0))
	{
		switch(msg.message)
		{
		case MM_WOM_DONE:
			dw->BlockDone((WAVEHDR*) msg.lParam);
			break;
		}
	}
	return 0;
}
